/*
NSIM is a simulator for the processor used in the HP-41 (Nut) and in the HP
Series 10 (Voyager) calculators.

$Id: rom.c 59 2003-11-07 02:48:10Z eric $
Copyright 1995, 2003 Eric Smith <eric@brouhaha.com>

NSIM is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License version 2 as published by the Free
Software Foundation.  Note that I am not granting permission to redistribute
or modify NSIM under the terms of any later version of the General Public
License.

This program is distributed in the hope that it will be useful (or at least
amusing), but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along with
this program (in the file "COPYING"); if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "nsim.h"
#include "rom.h"

#ifndef NULL
#define NULL ((void *) 0)
#endif

romword *ucode [MAX_PAGE][MAX_BANK];
bool *bpt      [MAX_PAGE][MAX_BANK];
char **source  [MAX_PAGE][MAX_BANK];
int active_bank[MAX_PAGE];
int hwpagebank [MAX_PAGE][MAX_BANK];
char *descpagebank [MAX_PAGE][MAX_BANK];
int movedfrom  [MAX_PAGE];

void init_rom (void)
{
  int b, p;
  for (p = 0; p < MAX_PAGE; p++)
    {
      active_bank [p] = 0;  /* all ROMs start with bank 0 selected */
      movedfrom [p] = 0;    /* page not moved */
      for (b = 0; b < MAX_BANK; b++)
	{
	  ucode  [p][b] = NULL;
	  bpt    [p][b] = NULL;
	  source [p][b] = NULL;
          hwpagebank [p][b] = HW_NONE;
	  descpagebank [p][b]= NULL;
	}
    }
}

/*static void alloc_page (int page, int bank)*/
void register_hwpage(int page, int bank, int hw, char*desc)
{
  int o;

  printf("Allocation of new page %1X, bank %1X, HW %1X, desc %s\n",
	 page, bank, hw, (desc ? desc : ""));
  ucode  [page][bank] = (romword *) calloc (PAGE_SIZE, sizeof (romword));
  bpt    [page][bank] = (bool *) calloc (PAGE_SIZE, sizeof (bool));
  source [page][bank] = (char **) calloc (PAGE_SIZE, sizeof (char *));

  if (! (ucode [page][bank] &&
	 bpt [page][bank] &&
	 source [page][bank]) )
    fatal (2, "memory allocation failed\n");

  hwpagebank[page][bank] = hw;
  if(desc)
     {
     descpagebank [page][bank]=(char*)malloc(strlen(desc)+1);
     if(!descpagebank[page][bank])
	 fatal (2, "desc memory allocation failed\n");
     strcpy(descpagebank[page][bank],desc);
     }
  for (o = 0; o < PAGE_SIZE; o++)
    bpt [page][bank][o] = 1;
}

void set_ucode (address a, int b, romword val)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  if( ! hwpagebank[p][b] )
    {
    printf("set_ucode():Unregistered HW page... %X %04X\n", b, p);
    register_hwpage(p, b, HW_ROM, "Defined by ucode");
    }
  ucode [p][b][a % PAGE_SIZE] = val;
}

void set_bpt    (address a, int b, bool val)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  if( ! hwpagebank[p][b] )
    {
    printf("set_btp():Unregistered HW page... %X %04X\n", b, p);
    register_hwpage(p, b, HW_ROM, "Defined by btp");
    }
  bpt [p][b][a % PAGE_SIZE] = val;
}

void set_source (address a, int b, char *val)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  if( ! hwpagebank[p][b] )
    {
    printf("set_source():Unregistered HW page... %X %04X\n", b, p);
    register_hwpage(p, b, HW_ROM, "Defined by source");
    }
  source [p][b][a % PAGE_SIZE] = val;
}

int get_hwpage(int p, int b)
{
    return hwpagebank[p][b];
}

romword get_ucode (address a)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  romword *page = ucode [p][active_bank [p]];
  if (hwpagebank[p][active_bank[p]])
    return (page [a % PAGE_SIZE]);
  else
    return (0);
}

bool get_bpt (address a)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  bool *page = bpt [p][active_bank [p]];
  if (hwpagebank[p][active_bank[p]])
    return (page [a % PAGE_SIZE]);
  else
    return (1);
}

char *get_source (address a)
{
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  char **page = source [p][active_bank [p]];
  if (hwpagebank[p][active_bank[p]])
    return (page [a % PAGE_SIZE]);
  else
    return (NULL);
}

void reset_bank( void )
{
  int ip;
  for( ip=0; ip<MAX_PAGE; ip++)
     active_bank[ip]=0; /* Reset to bank 0! */
}

void select_bank (address a, int b)
{
  /* change active bank only if requested bank actually exists */
  int p = a / PAGE_SIZE;  /* counting on the compiler to do this efficiently */
  if (hwpagebank[p&0xe][b] || hwpagebank[p|0x1][b])
    {
    active_bank [p&0xe] = (hwpagebank[p&0xe][b]?b:0);
    active_bank [p|0x1] = (hwpagebank[p|0x1][b]?b:0);
    }
  else if( hwpagebank[5][1] && (3==p || 5==p) )
    {
    active_bank [5] = b;
    }
  else
    printf("No page %X bank %X HW %X <== %04x\n",
	   p, b, hwpagebank[p][b], a);
}

int get_active_bank( int p )
{
    return( active_bank[p] );
}

void write_page( FILE *fhr, int p, int b )
{
   int ia;
   for( ia = 0; ia <PAGE_SIZE; ia++ )
      fprintf(fhr, "%04X %03X\n", ia, ucode [p][b][ia]);
}

void move_page( int sp, int dp )
{
   int ib;
   if(movedfrom[dp])
      {
      printf("PAGE %X ALREADY MOVED FROM %X\n", dp, sp);
      return;
      }
   for( ib=0; ib<MAX_BANK; ib++)
      {
      /* destination page */
      hwpagebank[dp][ib]= hwpagebank[sp][ib];
      active_bank[dp]=active_bank[sp];
      ucode[dp][ib]=ucode[sp][ib];
      source[dp][ib]=source[sp][ib];
      bpt[dp][ib]=bpt[sp][ib];
      descpagebank[dp][ib]=descpagebank[sp][ib];
      movedfrom[dp]=(movedfrom[sp]?movedfrom[sp]:sp);
      /* source page */
      hwpagebank[sp][ib]=HW_NONE;
      active_bank[sp]=0;
      ucode[sp][ib]=NULL;
      source[sp][ib]=NULL;
      bpt[sp][ib]=NULL;
      descpagebank[sp][ib]=NULL;
      movedfrom[sp]=0;
      }
}

void reset_moved_page(void)
{
   int ndp, dp, ip;
   do
      {
      for( ndp=0,ip=0; ip<MAX_PAGE; ip++)
         {
         if( (dp=movedfrom[ip]) )
	    {
	    if(movedfrom[dp])
	       ndp=movedfrom[dp];
	    else
	       {
	       int ib;
               for( ib=0; ib<MAX_BANK; ib++)
                  {
                  /* original location page */
                  hwpagebank[dp][ib]= hwpagebank[ip][ib];
                  active_bank[dp]=active_bank[ip];
                  ucode[dp][ib]=ucode[ip][ib];
                  source[dp][ib]=source[ip][ib];
                  bpt[dp][ib]=bpt[ip][ib];
                  descpagebank[dp][ib]=descpagebank[ip][ib];
                  movedfrom[dp]=0;
                  /* moved page */
                  hwpagebank[ip][ib]=HW_NONE;
                  active_bank[ip]=0;
                  ucode[ip][ib]=NULL;
                  source[ip][ib]=NULL;
                  bpt[ip][ib]=NULL;
                  descpagebank[ip][ib]=NULL;
                  movedfrom[ip]=0;
                  }
	       }
	    }
         }
      } while( ndp!=0 );
}

void show_pages(void)
{
   int ip, ib;

   for( ip=0; ip<MAX_PAGE; ip++)
      {
      printf("PAGE %X", ip);
      for( ib=0; ib<MAX_BANK; ib++)
	 if( hwpagebank[ip][ib] )
	    printf(" : %X %X %-10s",ib,hwpagebank[ip][ib],descpagebank[ip][ib]);
      printf("\n");
      }
   for( ip=0; ip<MAX_PAGE; ip++)
      if( movedfrom[ip] )
	  printf("PAGE %X moved from %X\n", ip, movedfrom[ip]);
}
